""" This is the entrypoint for ingestion launching """
import abc
import json
import logging
import sys
import tarfile
from pathlib import Path

# pylint: disable=E0401, R0903, R1721
from typing import Tuple, List

import pendulum
from pendulum import DateTime

from ingest_envoy.manifest_components import (
    INGESTION_ARTIFACTS_NAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    JSON,
    IngestionManifestKey,
    ManifestComponentIF,
    InputScienceProduct,
    InputGroup,
    ManifestParameters,
    OutputScienceProduct,
    AncillaryProduct,
    OutputGroup,
    MANIFEST_FILENAME,
    ParamsKey,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import (
    ScienceProductType,
    Telescope,
    IngestionManifestException,
    AncillaryProductType,
    find_output_tars,
)
from ingest_envoy.std_img_manifest_utils import ImageIngestionProductsFinder

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))

# pylint: disable=R0902, R0913


class ManifestIF(ManifestComponentIF):
    """Interface for all ingestion manifests"""

    def __init__(
        self,
        telescope: Telescope,
        sp_type: ScienceProductType,
        staging_source_dir: Path,
        locator: str,
        # all except EVLA_EB and VLASS catalog manifest have input group
        input_group: InputGroup,
        # every manifest has at least one output group
        output_group: OutputGroup,
        # image manifest has this:
        additional_metadata: AbstractTextFile = None,
    ):
        self.staging_source_dir = staging_source_dir
        self.sp_type = sp_type
        self.locator = locator
        self.input_group = input_group
        self.output_group = output_group
        self.telescope = telescope

        self.parameters = self.build_ingest_parameters(additional_metadata)

        self.files_found = [file for file in self.staging_source_dir.iterdir()]

    def build_ingest_parameters(self, additional_metadata: AbstractTextFile):
        """
        Make the "parameters" section of the manifest

        :return:
        """
        if self.sp_type not in [ScienceProductType.EVLA_CAL, ScienceProductType.IMAGE]:
            raise NotImplementedError()

        if additional_metadata:
            params = ManifestParameters(
                telescope=self.telescope,
                reingest=False,
                ngas_ingest=False,
                calibrate=False,
                staging_source_dir=self.staging_source_dir,
                additional_metadata=additional_metadata,
            )
        else:
            params = ManifestParameters(
                telescope=self.telescope,
                reingest=False,
                ngas_ingest=False,
                calibrate=False,
                staging_source_dir=self.staging_source_dir,
            )

        return params

    @abc.abstractmethod
    def write(self):
        """
        Write this manifest to a file, along with the artifacts tar and any other files required
        for this type of ingestion, at the ingest_path

        :param: location of files to be ingested, which is where we'll put the manifest
        :return:
        """
        raise NotImplementedError

    def __eq__(self, other):
        if isinstance(other, IngestionManifest):
            return (
                other.parameters == self.parameters
                and other.input_group == self.input_group
                and other.output_group == self.output_group
            )

        return False

    @abc.abstractmethod
    def to_json(self) -> JSON:
        """
        Turn this object into a JSON string suitable for writing to a file

        :return:
        """
        raise NotImplementedError


class IngestionManifestBuilder:
    """Builds ingestion manifest and associated files from files in ingestion_path"""

    def __init__(
        self,
        staging_source_dir: Path,
        sp_type: ScienceProductType,
        locator: str,
        telescope: Telescope,
        additional_metadata: AbstractTextFile = None,
    ):
        self.telescope = telescope
        self.staging_source_dir = staging_source_dir

        if sp_type not in [ScienceProductType.EVLA_CAL, ScienceProductType.IMAGE]:
            raise NotImplementedError(f"Don't know yet how to build a {sp_type.value} manifest")
        self.sp_type = ScienceProductType(sp_type)

        self.locator = locator

        # (for EVLA CAL, this will be None)
        self.additional_metadata = additional_metadata

        # self.files_found = [file for file in staging_source_dir.iterdir()]
        # if len(self.files_found) == 0:
        #     raise IngestionManifestException(f"No ingestion files found at {staging_source_dir}")

    def build(self) -> Tuple[ManifestIF, Path]:
        """
        Using only -relevant- files in ingestion_path, write the manifest
        and produce other files required for ingestion.

        :return: the ingestion manifest and the file containing its JSON
        """

        # # create any other ingestion files needed for this type of ingestion
        # self._find_additional_ingestion_files()

        if self.sp_type == ScienceProductType.EVLA_CAL:
            return self._build_evla_cal_manifest()

        return self._build_image_manifest()

        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=self.locator,
            sp_type=self.sp_type,
            staging_source_dir=self.staging_source_dir,
            input_group=self._build_input_group(),
            output_group=self._build_output_group(),
        )
        if self.additional_metadata:
            manifest.parameters.additional_metadata = self.additional_metadata

        manifest_file = manifest.write()

        self.write_ingestion_artifacts_tar()

        return manifest, manifest_file

    def _build_evla_cal_manifest(self):
        # create the manifest
        manifest = IngestionManifest(
            telescope=self.telescope,
            locator=self.locator,
            sp_type=self.sp_type,
            staging_source_dir=self.staging_source_dir,
            input_group=self._build_input_group(),
            output_group=self._build_output_group(),
        )
        manifest_file = manifest.write()
        self.write_ingestion_artifacts_tar()

        return manifest, manifest_file

    def _build_image_manifest(self):
        """
        Image manifest has additional_metadata, and output group is way more complicated
        :return:
        """
        # TODO:

    def _build_input_group(self):
        """
        Create the input group using the parameters.

        :return:
        """

        # N.B. this is sufficient for most types of ingestion,
        # but ALMA CALs will have multiple EB SPs, identified only by locator,
        # and VLBAs have no input group at all.
        sp_in = InputScienceProduct(locator=self.locator)

        return InputGroup([sp_in])

    def _define_output_science_products(self) -> List[OutputScienceProduct]:
        """
        Find in the staging dir a science product and, if applicable, its ancillary products.

        HEADS UP! ASSUMPTION: only one science product in the staging dir.
        This works for EVLA CAL and image ingestion but may need an overhaul
        for future ingestion types. We return a list because that's what there is
        in our example image ingestion manifest. YMMV, void where prohibited,
        professional driver on closed course, not FDIC insured.

        :return: output science products found
        """

        if self.sp_type == ScienceProductType.EVLA_CAL:
            tars_found = find_output_tars(self.files_found, self.staging_source_dir)
            for file in tars_found:
                sci_prod = OutputScienceProduct(type=self.sp_type, filename=str(file))
                return [sci_prod]

        elif self.sp_type == ScienceProductType.IMAGE:
            products_finder = ImageIngestionProductsFinder(self.staging_source_dir)
            science_products = products_finder.science_products
            ancillary_products = products_finder.ancillary_products

        #     image_products = self._find_image_products()
        #     sps = []
        #     for ip in image_products:
        #         if ip.type == AncillaryProductType.QUICKLOOK_IMAGE:
        #             # this is the science product, a quicklook image
        #             sp_itself = ip
        #         elif ip.type == AncillaryProductType.QUICKLOOK_RMS_IMAGE:
        #             sps.append(ip)
        #         elif ip.type == AncillaryProductType.PIPELINE_WEBLOG_TYPE:
        #             sps.append(ip)
        #         # elif ip.type == AncillaryProductType.
        #
        # aips = []
        # for file in tars_found:
        #     ap = self._build_ancillary_image_science_product(file)
        #     if ap:
        #         aips.append(ap)
        #
        #     maybe_weblogs = [file for file in self.staging_source_dir.glob(WEBLOG_FILENAME)]
        #     if len(maybe_weblogs) > 0:
        #         weblog = Path(maybe_weblogs[0])
        #         weblog_ap = AncillaryProduct(
        #             type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=str(weblog)
        #         )
        #         aips.append(weblog_ap)
        #
        #     sci_prod = OutputScienceProduct(
        #         type=sp_itself.type, filename=sp_itself.filename, ancillary_products=aips
        #     )
        #     return [sci_prod]
        # else:
        #     raise ValueError(f"Don't know yet how to handle a {self.sp_type.vaue}")

    # def _build_evla_cal_output_group(self) -> OutputGroup:
    #     # find science products
    #     science_products = self._define_output_science_products()
    #
    #     return OutputGroup(science_products=science_products)

    def _build_imaging_output_group(self) -> OutputGroup:
        """
        Create the output group using the parameters.

        :return:
        """

        products_finder = ImageIngestionProductsFinder(self.staging_source_dir)
        science_products = products_finder.science_products
        ancillary_products = products_finder.ancillary_products

        return OutputGroup(science_products=science_products, ancillary_products=ancillary_products)

    # def _build_ancillary_image_science_product(self, file: Path):
    #     """
    #     Image science products will have ancillary products of their very own,
    #     distinct from other ancillary products that might be in the staging dir.
    #
    #     :param file: a possible ancillary image product
    #     :return: the corresponding AncillaryProduct, if applicable
    #     """
    #     filename = str(file)
    #     if "image" in file.name:
    #         if file.name.endswith(".png"):
    #             return AncillaryProduct(type=AncillaryProductType.THUMBNAIL_IMG, filename=filename)
    #         elif file.name.endswith(".fits"):
    #             if "rms" in file.name:
    #                 return AncillaryProduct(
    #                     type=AncillaryProductType.QUICKLOOK_RMS_IMAGE, filename=filename
    #                 )
    #             else:
    #                 return AncillaryProduct(
    #                     type=AncillaryProductType.QUICKLOOK_IMAGE, filename=filename
    #                 )
    #
    # def _build_ancillary_product(self, file: Path) -> AncillaryProduct:
    #     """
    #     If this file is required for ingestion manifest creation,
    #     create an ancillary product from it.
    #
    #     :param file: file found in staging dir
    #     :return: ancillary product represented by this file, if any
    #     """
    #     if file.name == WEBLOG_FILENAME:
    #         return AncillaryProduct(
    #             type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=str(file)
    #         )
    #
    #     if AncillaryProductType.PIPELINE_ARTIFACTS.value in file.name:
    #         return AncillaryProduct(
    #             type=AncillaryProductType.PIPELINE_ARTIFACTS, filename=str(file)
    #         )
    #
    #     if AncillaryProductType.INGESTION_ARTIFACTS.value in file.name:
    #         return AncillaryProduct(
    #             type=AncillaryProductType.INGESTION_ARTIFACTS, filename=str(file)
    #         )
    #
    #     # this is not an ancillary product
    #     return None

    @staticmethod
    def build_artifacts_filename() -> str:
        """
        Build unique manifest filename in standard format.

        :return: the filename
        """
        current_time = pendulum.now()
        timestamp = format_timestamp(current_time)
        return f"{INGESTION_ARTIFACTS_NAME}{timestamp}{TARFILE_EXT}"

    def write_ingestion_artifacts_tar(self) -> Path:
        """
        Take the list of files and build a tar for inclusion into the archive.
        This happens in the staging area for ingestion.
        The EVLA CAL tar will contain just the manifest.

        :return: a .tar archive of the ingestion artifacts
        """
        ingestion_files = [file for file in self.staging_source_dir.iterdir() if file.is_file]
        manifest_file = find_manifest(self.staging_source_dir)
        ing_tar = self.staging_source_dir / self.build_artifacts_filename()
        with tarfile.open(ing_tar, "w") as ingestion_artifacts_tar:
            for file in ingestion_files:
                ingestion_artifacts_tar.add(file)

            # include the manifest
            if manifest_file not in ingestion_files:
                ingestion_artifacts_tar.add(manifest_file)

        return ing_tar

    # def _find_image_products(self) -> List[AncillaryProduct]:
    #     """
    #     Get ancillary products that belong to science products
    #
    #     :return:
    #     """
    #     ingestion_files = self._find_ingestion_files()
    #     aips = []
    #     for file in ingestion_files:
    #         aip = self._build_ancillary_image_science_product(file)
    #         if aip:
    #             aips.append(aip)
    #
    #     return aips

    # def _find_ancillary_products_for_img_ingest(self) -> List[AncillaryProduct]:
    #     """
    #     Round up any ancillary files found in image ingestion staging dir
    #
    #     :return: ancillary product(s) found
    #     """
    #
    #     ancillary_products = []
    #
    #     # TODO: START HERE THU 2021-07-29: the image files are science product ancillaries; the tars and the weblog are
    #     #  ancillary products. in test_creates_expected_manifest(), we expect 1 sci prod w/2 sp ancillaries,
    #     #  and 3 ancillary prods.
    #
    #     ingestion_files = self._find_ingestion_files()
    #     for file in ingestion_files:
    #         maybe_ap = self._build_ancillary_product(file)
    #         if maybe_ap and maybe_ap not in ancillary_products:
    #             ancillary_products.append(maybe_ap)
    #
    #     # There should be a weblog in here; grab it
    #     maybe_weblogs = [file for file in self.files_found if file.name == WEBLOG_FILENAME]
    #     if len(maybe_weblogs) > 0:
    #         file = self.staging_source_dir / WEBLOG_FILENAME
    #         weblog_ap = AncillaryProduct(
    #             type=AncillaryProductType.PIPELINE_WEBLOG_TYPE, filename=str(file)
    #         )
    #         if weblog_ap not in ancillary_products:
    #             ancillary_products.append(weblog_ap)
    #     else:
    #         raise FileNotFoundError(f"No weblog found in {self.staging_source_dir}")
    #
    #     return ancillary_products

    def _find_ingestion_files(self) -> List[Path]:
        """
        Round up any other necessary ingestion file(s)

        :return: additional relevant files found in ingestion path, if any
        """

        if self.sp_type == ScienceProductType.EVLA_CAL:
            # there won't be any others
            return []

        if self.sp_type == ScienceProductType.IMAGE:
            return [file for file in self.staging_source_dir.iterdir()]

        # TODO when the time comes: we'll have extra information for other ingestion types;
        #  see archive-metaproject
        raise NotImplementedError


class IngestionManifest(ManifestIF):
    """write ingestion manifest to file"""

    def ingestion_path(self) -> Path:
        return self.parameters.staging_source_dir

    def write(self) -> Path:
        """
        Write the manifest .json file.

        :return:
        """

        me_dict = self.to_json()
        output_path = self.staging_source_dir / MANIFEST_FILENAME

        to_write = json.dumps(me_dict, indent=4)
        with open(output_path, "w") as out:
            out.write(to_write)

        return output_path

    def to_json(self) -> JSON:
        """
        Turn this object into a JSON string suitable for writing to a file

        :return:
        """

        me_dict = dict(self.__dict__)

        to_return = {
            IngestionManifestKey.PARAMETERS.value: self.parameters.to_json(),
            IngestionManifestKey.INPUT_GROUP.value: me_dict[
                IngestionManifestKey.INPUT_GROUP.value
            ].to_json(),
            IngestionManifestKey.OUTPUT_GROUP.value: me_dict[
                IngestionManifestKey.OUTPUT_GROUP.value
            ].to_json(),
        }
        if self.parameters.additional_metadata:
            to_return[ParamsKey.ADDITIONAL_METADATA.value] = str(
                self.parameters.additional_metadata
            )

        return to_return


def format_timestamp(datetime: DateTime) -> str:
    """
    Format the current time as
        2021_07_01T13_49_17.237

    :param datetime: current timestamp
    :return: timestamp suitable for ingestion manifest filename
    """

    return datetime.format("YYYY_MM_DDThh_mm_ss.SSS")


def find_manifest(ingestion_path: Path) -> Path:
    """
    Find the ingestion manifest at this ingestion path.

    :param ingestion_path: home of ingestion files
    :return:
    """
    for json_file in ingestion_path.glob(MANIFEST_FILENAME):
        return json_file

    raise FileNotFoundError(f"No ingestion manifest found at {ingestion_path}")
