#!/usr/bin/env bash
#
#  A file to investigate a directory and create an imaging
#  artifacts tar file, then stage files for ingestion
#  into NGAS. Adapted from AAT-PPI for Workspaces use
#
#  Arguments:
#       1: The workflow directory name
#       2: The staging directory root path
#       3: A name for the artifacts file
#
set -o errexit -o nounset -o xtrace


#
# command line arguments
#
# Expanded for flexibility across processing sites, and also
# allowing for the decoupling of the artifact's filename from
# the working directory.
#
# Because of the potential, large differences in directory
# layout between the VLASS cache, internal processing, and
# external requests, this removes any necessity for the
# script to make assumptions about paths, etc.
#
WORKFLOW_DIR=$1;shift
STAGE_DIR=$1;shift
FILENAME=$1;shift

SPOOL_DIR=$(./pycapo -P "${CAPO_PROFILE}" -q edu.nrao.workspaces.ProcessingSettings.rootDirectory)
SOURCE_DIR=${SPOOL_DIR}/${WORKFLOW_DIR}/products


# Create the staging directory carefully:
mkdir -p ${STAGE_DIR}/${WORKFLOW_DIR}


#
# Check for the state of the weblog:  If we've got an unzipped weblog directory, we
# should bundle it up (in preference to the zipped one) in case it's been updated
# and edited by the DAs.  If there isn't one, check for the tgz file.
#
# However happens, we want the weblog copied/moved to the staging area separate from
# the rest of the pipeline artifacts (weblogs are important enough to be defined as
# an ancillary product in their own right).
#
WEBLOG_DIR=$(ls -t ${SOURCE_DIR} | grep pipeline- | head -1)
if [[ -n "$WEBLOG_DIR" ]]
then
    #
    # Remove all the old weblogs before we re-tar the open one:
    #
    for zipped_log in $(ls -t ${SOURCE_DIR}/*weblog.tgz);do rm ${zipped_log};done


    # Tar & Zip the weblog
    tar -C${SOURCE_DIR} -czf ${SOURCE_DIR}/weblog.tgz ${WEBLOG_DIR}

    if [[ $? -ne 0 ]]
    then
        echo "Creation of weblog.tgz failed, exiting"
        exit -1
    fi
fi
# The weblog may have a prefix, so account for that:
/bin/cp -f ${SOURCE_DIR}/*weblog.tgz ${STAGE_DIR}/${WORKFLOW_DIR}
# NOTE: This fails if we have no weblog, previously we allowed that
#       possibility.

#
# Prepare the artifacts file, we'll collect everything to be ingested in a directory.
#
if [[ -e ${STAGE_DIR}/${WORKFLOW_DIR}/${FILENAME}.tar ]]
then
    echo "Artifacts Tar File Already Exists!  Removing the file for recreation"
    /bin/rm -f ${STAGE_DIR}/${WORKFLOW_DIR}/${FILENAME}.tar
fi

# tar a subset of files into a tar archive in the storage path
#   Skipped: FITS files, Weblog (zipped and unzipped)
#
tar --exclude=\*.fits --exclude=pipeline-\* --exclude=*weblog.tgz -C${SOURCE_DIR} -cvf ${STAGE_DIR}/${WORKFLOW_DIR}/${FILENAME}.tar .
if [[ $? -ne 0 ]]
then
    echo "Creation of pipeline artifacts tar file failed, exiting"
    exit -1
fi


# Now link the FITS files over to the staging directory (Create hard link, and be insistent)
#
# We just want the file name, so we'll work in the source directory
pushd ${SOURCE_DIR}
# link the files
for fitsFile in $(ls -t *.fits);do ln -f ${SOURCE_DIR}/${fitsFile} ${STAGE_DIR}/${WORKFLOW_DIR};done
# move back to the working directory
popd
